@@ -11,35 +11,49 @@
  * the author assume any responsibility for damages incurred with its 
  * use.
  *
+ *	$Id: xbatt.c,v 1.2 1998/09/11 15:43:16 eto Exp eto $
  */
 
 /*
  * HISTORY
- *	95/03/27	0.1	XBM version
- *	95/03/28	0.2	refine bitmap management
- *	95/03/30	0.3	use XPM library
- *	95/04/01	0.3.1	by Noriyuki Takahashi (nor@aecl.ntt.jp)
+ *	95/03/27  0.1	XBM version
+ *	95/03/28  0.2	refine bitmap management
+ *	95/03/30  0.3	use XPM library
+ *	95/04/01  0.3.1	by Noriyuki Takahashi (nor@aecl.ntt.jp)
  *		Display plug and charging status.
  *		Support gray-scale and mono (compile-time option).
- *	95/04/03	0.4	Support gray-scale and mono (auto detect).
- *				Update status, when catch key event.
- *	95/04/19	1.0	Official release with some documents!!!!
- *	95/04/27	1.1	Support Linux apm-bios 0.5
+ *	95/04/03  0.4	Support gray-scale and mono (auto detect).
+ *			Update status, when catch key event.
+ *	95/04/19  1.0	Official release with some documents!!!!
+ *	95/04/27  1.1	Support Linux apm-bios 0.5
  *				by pcs00294@asciinet.or.jp (Koyama Tadayoshi)
- *	97/01/01	1.2b1	Modify some place
- *	97/01/17	1.2b3	Support Linux-1.3.58 or lator
- *					by hisama@bunff.trad.pfu.co.jp
- *					   mac@pssys.flab.fujitsu.co.jp
- *					   yoshi@funival.mfd.cs.fujitsu.co.jp
- *				Support for PC which does not return
- *				buttery life (e.g. ThinkPad 535)
- *					by umemoto
- *				Add check timer routine for resume problem
- *				Show battery life, when the mouse
- *				cursor located on battery icon.
- *	97/01/20	1.2b4	Changes for Xt related part (e.g. resources)
- *					by mi@sy.ssl.fujitsu.co.jp
- *				Change small bugs
+ *	97/01/01  1.2b1	Modify some place
+ *	97/01/17  1.2b3	Support Linux-1.3.58 or lator
+ *				by hisama@bunff.trad.pfu.co.jp
+ *				   mac@pssys.flab.fujitsu.co.jp
+ *				   yoshi@funival.mfd.cs.fujitsu.co.jp
+ *			Support for PC which does not return
+ *			buttery life (e.g. ThinkPad 535)
+ *				by umemoto
+ *			Add check timer routine for resume problem
+ *			Show battery life, when the mouse
+ *			cursor located on battery icon.
+ *	97/01/20  1.2b4	Changes for Xt related part (e.g. resources)
+ *				by mi@sy.ssl.fujitsu.co.jp
+ *			Fix minor bugs
+ *	97/03/07  1.2	Release
+ *	98/09/11  1.2.1	Change error message
+ *				by Kenneth Ingham <ingham@i-pi.com>
+ *	98/09/??  1.3	NetBSD support
+ *				by Tatoku Ogaito (tacha@trap.fukui-med.ac.jp)
+ *			Modify battery picture
+ *				by Takeshi OHASHI (ohashi@ai.kyutech.ac.jp)
+ *			Support new command line option:
+ *			    -small, -geometry, -width, -height
+ *				by Takeshi OHASHI (ohashi@ai.kyutech.ac.jp)
+ *				 & mrt@mickey.ai.kyutech.ac.jp
+ *			WindowMaker Dock (-withdrawn) patch
+ *				by Takao KAWAMURA (kawamura@debian.or.jp)
  */
 
 #include <stdio.h>
@@ -65,8 +79,15 @@
 # define APMDEV22	"/dev/apm"
 #endif
 
+#ifdef __NetBSD__
+# include <machine/apmvar.h>
+#endif
+
+
 #include "pixmaps/battery.xpm"
 #include "pixmaps/unknown.xpm"
+#include "pixmaps/small-battery.xpm"
+#include "pixmaps/small-unknown.xpm"
 #include "bitmaps/full.xbm"
 #include "bitmaps/digit0.xbm"
 #include "bitmaps/digit1.xbm"
@@ -113,21 +134,24 @@ static	struct status	getBatteryStatus();
 static	void		updateWindow(struct status);
 static	void		initWindowResource();
 
-#define	BatteryWidth	39
-#define	BatteryHeight	39
+#define	BatteryWidth		39
+#define	BatteryHeight		39
+#define	BatterySmallWidth	16
+#define	BatterySmallHeight	16
 
 static	void setColorSymbol(unsigned int, unsigned int, unsigned int);
 static	void goOut();
 static	void CallbackTyped(Widget, char*, XEvent*, Boolean*);
 static	void CallbackEnter(Widget, char*, XEvent*, Boolean*);
 static	void CallbackLeave(Widget, char*, XEvent*, Boolean*);
+static	void CallbackResize(Widget, char*, XEvent*, Boolean*);
 
 static	time_t		lastUpdateTime = 0;
 static	int		showStatus = 0;
 static	int		forceRedraw = 0;
 static	int		apmfd = 0;
-static	Widget		toplevel;
-static	Pixmap		xpmData, xpmMask;
+static	Widget		toplevel, icon;
+static	Pixmap		xpmData, xpmMask, xpmDataPadded;
 static	GC		gc;
 static	XpmAttributes	attr;
 static	XtIntervalId	timer = 0;
@@ -138,9 +162,15 @@ static	int		displayType;	/* Mono, Gray, Color */
 #define	MAXSYM		30
 static	XpmColorSymbol	sym[MAXSYM];
 static	unsigned int	nsym = 0;
+static	u_int		windowHeight;
+static	u_int		windowWidth;
+static	u_int		xpadding;
+static	u_int		ypadding;
 
 typedef struct _SetupRec{
-    int	timerInterval;
+    int		timerInterval;
+    Boolean	small;
+    Boolean	withdrawn;
 } SetupRec, *SetupPtr;
 
 static SetupRec appResources;
@@ -149,11 +179,22 @@ static XtResource resources[] = {
     {"timerInterval", "TimerInterval", XtRInt, sizeof(int),
      XtOffsetOf(struct _SetupRec, timerInterval), XtRImmediate,
      (XtPointer)UPDATE_INTERVAL},
+    {"small", "Small", XtRBoolean, sizeof(Boolean),
+     XtOffsetOf(struct _SetupRec, small), XtRImmediate,
+     (XtPointer) FALSE},
+    {"withdrawn", "Withdrawn", XtRBoolean, sizeof(Boolean),
+     XtOffsetOf(struct _SetupRec, withdrawn), XtRImmediate,
+     (XtPointer) FALSE},
 };
 
 static XrmOptionDescRec	options[] = {
     {"-timerInterval", ".timerInterval", XrmoptionSepArg,
      (XtPointer)UPDATE_INTERVAL},
+    {"-small", ".small", XrmoptionNoArg, "TRUE"},
+    {"-geometry", ".geometry", XrmoptionSepArg, (XtPointer) NULL},
+    {"-width", ".width", XrmoptionSepArg, (XtPointer) NULL},
+    {"-height", ".height", XrmoptionSepArg, (XtPointer) NULL},
+    {"-withdrawn", ".withdrawn", XrmoptionNoArg, "TRUE"},
 };
 
 struct	Digits {
@@ -175,14 +216,22 @@ struct Digits	digits[] = {
     {digit9_width, digit9_height, digit9_bits}
 };
 
+String fallback_resouces[] = {
+    "*width: 39",
+    "*height: 39",
+};
+
 main(
   int	argc,
   char	*argv[]
 ) {
     XSizeHints		size_hints;
+    XWMHints		wm_hints;
     Visual*	visual;
     int		depth;
+    String	geom;
     XtAppContext appContext;
+    Boolean	withdrawn = False;
 
 #if !defined(NOAPM) && defined(__FreeBSD__)
     /* initialize APM Interface */
@@ -195,11 +244,18 @@ main(
     }
 #endif
 
+#ifdef __NetBSD__
+    if ((apmfd = open("/dev/apm", O_RDONLY, 0755)) == -1) {
+	perror("xbatt(open device):");
+	exit(1);
+    }
+#endif
+
     /* start X-Window session */
     XtSetLanguageProc( NULL, NULL, NULL );
     toplevel = XtOpenApplication(&appContext, "XBatt",
 				 options, XtNumber(options),
-				 &argc, argv, NULL, 
+				 &argc, argv, fallback_resouces, 
 				 sessionShellWidgetClass, NULL, ZERO);
     if (!toplevel) {
 	fprintf(stderr, "xbatt: cannot start x-session\n");
@@ -207,10 +263,12 @@ main(
     }
     if (argc != 1) {
 	fprintf(stderr, "usage: xbatt [X-Toolkit standard option]\n");
+	fprintf(stderr, "             [-timerInterval usec] [-small]\n");
+	fprintf(stderr, "             [-withdrawn] [-width w] [-height h]\n");
 	exit(1);
     }
 
-    /* get rsource information */
+    /* get resource information */
     XtGetApplicationResources(toplevel, &appResources,
 			      resources, XtNumber(resources), NULL, 0);
 
@@ -239,17 +297,57 @@ main(
 	}
     }
 
+    /* get current property */
+    XtVaGetValues(toplevel,
+		  XtNwidth, &windowWidth,
+		  XtNheight, &windowHeight,
+		  XtNgeometry, &geom,
+		  NULL);
+
+    size_hints.flags = 0;
+    if (geom) {
+	int x, y;
+	XParseGeometry(geom, &x, &y, &windowWidth, &windowHeight);
+	size_hints.flags |= USPosition;
+	size_hints.x = x;
+	size_hints.y = y;
+    }
+
+    if (windowWidth < BatteryWidth ||
+        windowHeight < BatteryHeight) {
+	appResources.small = TRUE;
+        if (windowWidth < BatterySmallWidth ||
+            windowHeight < BatterySmallHeight) {
+            windowWidth = BatterySmallWidth;
+            windowHeight = BatterySmallHeight;
+        }
+    }
+
+    size_hints.flags |= PMinSize;
+    size_hints.min_height = BatterySmallHeight;
+    size_hints.min_width  = BatterySmallWidth;
+
     /* set toplevel property */
     XtVaSetValues(toplevel,
 		  XtNmappedWhenManaged, False,
 		  XtNinput, True,
-		  XtNwidth, BatteryWidth,
-		  XtNheight,BatteryHeight,
-		  XtNminWidth, BatteryWidth,
-		  XtNminHeight, BatteryHeight,
+		  XtNwidth, windowWidth,
+		  XtNheight, windowHeight,
+		  XtNminWidth, BatterySmallWidth,
+		  XtNminHeight, BatterySmallHeight,
 		  XtNiconName, "xbatt",
 		  NULL);
 
+    /* create icon widget */
+    if (appResources.withdrawn) {
+	icon = XtVaCreateManagedWidget("icon", sessionShellWidgetClass,
+				       toplevel, NULL);
+	if (!icon) {
+	    fprintf(stderr, "usage: xbatt [X-Toolkit standard option]\n");
+	    exit(1);
+	}
+    }
+
     /* realize toplevel widget */
     XtRealizeWidget(toplevel);
 
@@ -261,10 +359,32 @@ main(
 		      (XtEventHandler)CallbackEnter, NULL);
     XtAddEventHandler(toplevel, LeaveWindowMask, False,
 		      (XtEventHandler)CallbackLeave, NULL);
+    XtAddEventHandler(toplevel, ResizeRedirectMask, False,
+		      (XtEventHandler)CallbackResize, NULL);
+    if (appResources.withdrawn) {
+	XtAddEventHandler(icon, KeyPressMask, False,
+			  (XtEventHandler)CallbackTyped, NULL);
+	XtAddEventHandler(icon, EnterWindowMask, False,
+			  (XtEventHandler)CallbackEnter, NULL);
+	XtAddEventHandler(icon, LeaveWindowMask, False,
+			  (XtEventHandler)CallbackLeave, NULL);
+    }
+
+    /* set hint information */
+    XSetWMNormalHints(XtDisplay(toplevel), XtWindow(toplevel),
+                      &size_hints);
 
     /* display current battery status, and set timer callback */
     updateStatus(NULL, NULL);
 
+    /* specify initial state and icon window */
+    if (appResources.withdrawn) {
+	wm_hints.flags = StateHint | IconWindowHint;
+	wm_hints.icon_window = XtWindow(icon);
+	wm_hints.initial_state = WithdrawnState;
+	XSetWMHints(XtDisplay(toplevel), XtWindow(toplevel), &wm_hints);
+    }
+
     /* start main loop */
     XtAppMainLoop(appContext);
 
@@ -372,6 +492,17 @@ struct status getBatteryStatus()
     }
 #endif	/* FreeBSD */
 
+#ifdef __NetBSD__
+    struct apm_power_info info;
+    if( ioctl(apmfd, APM_IOC_GETPOWER, &info) == -1 ) {
+	fprintf(stderr, "xbatt: ioctl APM_IOC_GETPOWER failed\n");
+	exit(1);
+    }
+    ret.remain = info.battery_life;
+    ret.acline = info.ac_state;
+    ret.charge = info.battery_state;
+#endif /* __NetBSD__ */
+
 #ifdef	__linux__
     char	buffer[64];
     FILE*	fp;
@@ -458,6 +589,9 @@ void updateWindow(struct status s)
     if (xpmMask) {
 	XFreePixmap(XtDisplay(toplevel), xpmMask);
     }
+    if (xpmDataPadded) {
+	XFreePixmap(XtDisplay(toplevel), xpmDataPadded);
+    }
     XpmFreeAttributes(&attr);
 
     /* set attribute value mask */
@@ -480,15 +614,29 @@ void updateWindow(struct status s)
 
     /* create pixmap data */
     if (s.remain != APM_STAT_UNKNOWN) {
-	ret = XpmCreatePixmapFromData(XtDisplay(toplevel),
-				      XtWindow(toplevel),
-				      batt_xpm, &xpmData,
-				      &xpmMask, &attr);
+	if (!appResources.small) {
+	    ret = XpmCreatePixmapFromData(XtDisplay(toplevel),
+					  XtWindow(toplevel),
+					  batt_xpm, &xpmData,
+					  &xpmMask, &attr);
+	} else {
+	    ret = XpmCreatePixmapFromData(XtDisplay(toplevel),
+					  XtWindow(toplevel),
+					  small_batt_xpm, &xpmData,
+					  &xpmMask, &attr);
+	}
     } else {
-	ret = XpmCreatePixmapFromData(XtDisplay(toplevel),
-				      XtWindow(toplevel),
-				      ubatt_xpm, &xpmData,
-				      &xpmMask, &attr);
+	if (!appResources.small) {
+	    ret = XpmCreatePixmapFromData(XtDisplay(toplevel),
+					  XtWindow(toplevel),
+					  ubatt_xpm, &xpmData,
+					  &xpmMask, &attr);
+	} else {
+	    ret = XpmCreatePixmapFromData(XtDisplay(toplevel),
+					  XtWindow(toplevel),
+					  small_ubatt_xpm, &xpmData,
+					  &xpmMask, &attr);
+	}
     }
     switch (ret) {
       case XpmSuccess:
@@ -507,8 +655,11 @@ void updateWindow(struct status s)
 	break;
     }
 
+    xpadding = (windowWidth - attr.width) / 2;
+    ypadding = (windowHeight - attr.width) / 2;
+
     /* resize window size to pixmap size */
-    XtResizeWidget(toplevel, attr.width, attr.height, 1);
+    XtResizeWidget(toplevel, windowWidth, windowHeight, 1);
 
     /* if show status is active, then draw status info */
     if ((showStatus != 0) && (s.remain != APM_STAT_UNKNOWN)) {
@@ -517,11 +668,23 @@ void updateWindow(struct status s)
 	XSetForeground(XtDisplay(toplevel), gc,
 		       BlackPixel(XtDisplay(toplevel),
                                   DefaultScreen(XtDisplay(toplevel))));
-	XDrawRectangle(XtDisplay(toplevel), xpmData, gc, 17, 17, 14, 8);
+	if (!appResources.small) {
+	    XDrawRectangle(XtDisplay(toplevel), xpmData, gc,
+			   17, 17, 14, 8);
+	} else {
+	    XDrawRectangle(XtDisplay(toplevel), xpmData, gc,
+			   4, 4, 13, 8);
+	}
 	XSetForeground(XtDisplay(toplevel), gc,
 		       WhitePixel(XtDisplay(toplevel),
                                   DefaultScreen(XtDisplay(toplevel))));
-	XFillRectangle(XtDisplay(toplevel), xpmData, gc, 18, 18, 13, 7);
+	if (!appResources.small) {
+	    XFillRectangle(XtDisplay(toplevel), xpmData, gc,
+			   18, 18, 13, 7);
+	} else {
+	    XFillRectangle(XtDisplay(toplevel), xpmData, gc,
+			   5, 5, 11, 7);
+	}
 
 	if (s.remain == 100) {
 	    bm = XCreatePixmapFromBitmapData(XtDisplay(toplevel),
@@ -535,9 +698,15 @@ void updateWindow(struct status s)
 						 DefaultScreen(
 						     XtDisplay(toplevel))),
 					     1);	/* Depth */
-	    XCopyPlane(XtDisplay(toplevel), bm, xpmData, gc, 0, 0,
-		       full_width, full_height,
-		       19, 19, 1);
+	    if (!appResources.small) {
+		XCopyPlane(XtDisplay(toplevel), bm, xpmData, gc, 0, 0,
+			   full_width, full_height,
+			   19, 19, 1);
+	    } else {
+		XCopyPlane(XtDisplay(toplevel), bm, xpmData, gc, 0, 0,
+			   full_width, full_height,
+			   6, 6, 1);
+	    }
 	    XFreePixmap(XtDisplay(toplevel), bm);
 	} else {
 	    int	d = s.remain / 10;
@@ -553,8 +722,13 @@ void updateWindow(struct status s)
 						 DefaultScreen(
 						     XtDisplay(toplevel))),
 					     1);
-	    XCopyPlane(XtDisplay(toplevel), bm, xpmData, gc, 0, 0, 3, 5,
-		       19, 19, 1);
+	    if (!appResources.small) {
+		XCopyPlane(XtDisplay(toplevel), bm, xpmData, gc, 0, 0, 3, 5,
+			   19, 19, 1);
+	    } else {
+		XCopyPlane(XtDisplay(toplevel), bm, xpmData, gc, 0, 0, 3, 5,
+			   7, 6, 1);
+	    }
 	    XFreePixmap(XtDisplay(toplevel), bm);
 
 	    d = s.remain % 10;
@@ -570,41 +744,70 @@ void updateWindow(struct status s)
 						 DefaultScreen(
 						     XtDisplay(toplevel))),
 					     1);
-	    XCopyPlane(XtDisplay(toplevel), bm, xpmData, gc, 0, 0, 3, 5,
-		       23, 19, 1);
+	    if (!appResources.small) {
+		XCopyPlane(XtDisplay(toplevel), bm, xpmData, gc, 0, 0, 3, 5,
+			   23, 19, 1);
+	    } else {
+		XCopyPlane(XtDisplay(toplevel), bm, xpmData, gc, 0, 0, 3, 5,
+			   11, 6, 1);
+	    }
 	    XFreePixmap(XtDisplay(toplevel), bm);
 
-	    bm = XCreatePixmapFromBitmapData(XtDisplay(toplevel),
-					     XtWindow(toplevel),
-					     percent_bits,
-					     percent_width,
-					     percent_height,
-					     BlackPixel(XtDisplay(toplevel),
-						 DefaultScreen(
-						     XtDisplay(toplevel))),
-					     WhitePixel(XtDisplay(toplevel),
-						 DefaultScreen(
-						     XtDisplay(toplevel))),
-					     1);
-	    XCopyPlane(XtDisplay(toplevel), bm, xpmData, gc, 0, 0, 3, 5,
-		       27, 19, 1);
-	    XFreePixmap(XtDisplay(toplevel), bm);
+	    if (!appResources.small) {
+		bm = XCreatePixmapFromBitmapData(XtDisplay(toplevel),
+						 XtWindow(toplevel),
+						 percent_bits,
+						 percent_width,
+						 percent_height,
+						 BlackPixel(XtDisplay(toplevel),
+							    DefaultScreen(XtDisplay(toplevel))),
+						 WhitePixel(XtDisplay(toplevel),
+							    DefaultScreen(XtDisplay(toplevel))),
+						 1);
+		XCopyPlane(XtDisplay(toplevel), bm, xpmData, gc, 0, 0, 3, 5,
+			   27, 19, 1);
+		XFreePixmap(XtDisplay(toplevel), bm);
+	    }
 	}
     }
 
+    /* make padded pixmap */ 
+    xpmDataPadded = XCreatePixmap(XtDisplay(toplevel), XtWindow(toplevel),
+				  windowWidth,
+				  windowHeight,
+				  XDefaultDepth(XtDisplay(toplevel),
+						DefaultScreen(XtDisplay(toplevel))));
+
+    XCopyArea(XtDisplay(toplevel), xpmData, xpmDataPadded, gc, 0, 0,
+              attr.width, attr.height, xpadding, ypadding);
+
     /* set pixmap data */
     XSetWindowBackgroundPixmap(XtDisplay(toplevel),
 			       XtWindow(toplevel),
-			       xpmData);
+			       xpmDataPadded);
+    if (appResources.withdrawn) {
+	XSetWindowBackgroundPixmap(XtDisplay(toplevel),
+				   XtWindow(icon),
+				   xpmDataPadded);
+    }
 
     /* if pixmap has transparent pixel, set mask pixmap */
     if (xpmMask) {
 	XShapeCombineMask(XtDisplay(toplevel), XtWindow(toplevel),
-			  ShapeBounding, 0, 0, xpmMask, ShapeSet);
+			  ShapeBounding, xpadding, ypadding,
+			  xpmMask, ShapeSet);
+	if (appResources.withdrawn) {
+	    XShapeCombineMask(XtDisplay(toplevel), XtWindow(icon),
+			      ShapeBounding, xpadding, ypadding,
+			      xpmMask, ShapeSet);
+	}
     }
 
     /* redraw window */
     XClearWindow(XtDisplay(toplevel), XtWindow(toplevel));
+    if (appResources.withdrawn) {
+	XClearWindow(XtDisplay(toplevel), XtWindow(icon));
+    }
 
     /* map window */
     XtMapWidget(toplevel);
@@ -773,3 +976,37 @@ void CallbackLeave(
 
 /*    fprintf(stderr, "<<< Leave\n");*/
 }
+
+void CallbackResize(
+  Widget	widget,
+  char*		tag,
+  XEvent*	xe,
+  Boolean*	b
+) {
+    if (!XtIsRealized(widget)) {
+        return;
+    }
+
+    windowWidth = xe->xresizerequest.width;
+    windowHeight = xe->xresizerequest.height;
+
+    if ((windowWidth < BatteryWidth) || (windowHeight < BatteryHeight)) {
+	appResources.small = TRUE;
+        if ((windowWidth < BatterySmallWidth)
+	    || (windowHeight < BatterySmallHeight)) {
+            windowWidth = BatterySmallWidth;
+            windowHeight = BatterySmallHeight;
+        }
+    } else {
+	appResources.small = FALSE;
+    }
+
+    /* reset status flag */
+    showStatus = 0;
+
+    /* display current battery status, and set timer callback */
+    forceRedraw = 1;
+    updateStatus(NULL, NULL);
+
+/*    fprintf(stderr, "<<< Resize\n");*/
+}
